#ifndef TREE_H
#define TREE_H

// forward declcarations
template <typename T> class avl_t;
template <typename T> std::ostream& operator<<(std::ostream&, const avl_t<T>&);

template <typename T>
class avl_t {
  private:
  struct node_t        // an all public class with data only, no member ftns
  {
        T        data;
        node_t   *left;
        node_t   *right;
        int      bal;

        node_t(const T& d = T(), node_t *l = NULL,
                                 node_t *r = NULL, int b = 0) : \
                   data(d), left(l), right(r), bal(b) \
                                                    { };
  };

  public:
  // constructors (overloaded)
  avl_t();

  // copy constructor
//avl_t(const avl_t& rhs);

  // destructors
  ~avl_t()                             { clear(); }

  // friends -- note the extra <> telling the compiler to instantiate
  // a templated version of the operator<< -- <T> is also legal, i.e.,
  // friend std::ostream& operator<< <T>(std::ostream& s, const avl_t&);
  friend
  std::ostream& operator<< <>(std::ostream& s, const avl_t& rhs);
  friend
  std::ostream& operator<<(std::ostream& s, avl_t *rhs)
                                       { return(s << (*rhs)); }
  void inorder(std::ostream& s, node_t* const &t) const;

  // assignment operator
  const avl_t&        operator=(const avl_t&);

  // operators

  // members
  bool      empty() const              { return root == NULL ? true : false;}
  bool      contains(const T& x) const { return contains(x,root); }
  void      insert(const T& x)         { insert(x, root); }
  void      erase(const T& x)          { erase(x, root); }
  void      clear()                    { clear(root); }
  const T&  min() const                { if(!empty()) return(min(root)->data); }
  const T&  max() const                { if(!empty()) return(max(root)->data); }

  // private: only available to this class
  private:
  node_t    *root;

  bool      contains(const T&, node_t* ) const;
  int       insert(const T& x, node_t* &t);
  int       erase(const T& x, node_t* &t);
  void      clear(node_t* &);
  node_t*   min(node_t* ) const;
  node_t*   max(node_t* ) const;
  node_t*   clone(node_t* ) const;
  void      rotate_left(node_t* &t);
  void      rotate_right(node_t* &t);

};

#endif
